#!/usr/bin/perl -w

# Moblin Application Checker
# Journal Parser and Report Generator Module (report.pl)
#
# Copyright (C) 2007-2009 The Linux Foundation. All rights reserved.
#
# This program has been developed by ISP RAS for LF.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

use strict;

sub BEGIN {
	my $utils_path = $0;
	$utils_path =~ s!/[^/]+$!!;
	if ($utils_path !~ m!^/!) {
		my $cwd = `pwd`;
		$cwd =~ s/[\r\n]//g;
		$utils_path = "$cwd/$utils_path";
	}
	push @INC, $utils_path;
}

use Common;

autoflush_on();

################################################################################
# Global variables

# Paths
my $result_path;           # Directory containing all test results.
my $result_dir;            # Directory with the current test result.
my $result_archive;        # TAR.GZ archive name to pack the results into.
my $application_name;      # Name of the application tested.

# Host information
my $host_name;
my $host_os;
my $host_machine;
my $test_arch;
my $host_kernel;

# Test information
my $moblin_version;
my $test_comments;

# List of all journal titles
my @journal_titles = ();

# List of HTML journal anchors in the same order as they appear in the tree
my @journal_anchors_sorted = ();

# Information for each journal
# %journal_data: <journal-title> => {
#                    'JOURNAL_FILE' => <file-name>
#                    'TEST_TOOL' => <tool-name>
#                    'START_TIME' => <time>
#                    'IS_RESULT' => (1|0|undef),            # undef - no journal, 0 - journal empty
#                    'PROBLEMS_LIST' => [
#                        {
#                            'SEVERITY' => (Warning|Error|Notice|...),
#                            'TEST_CASE' => <test-case ID>,
#                            'SUBCATEGORY' => <SUBCAT-ID>,
#                            'ELEMENT' => <element>,
#                            'COMMENT' => <comment>,
#                            'DETAILS' => <description-text>
#                        },
#                        ...
#                    ],
#                    'STATISTICS' => {
#                        'PASSES' => <number>,
#                        'FAILS' => <number>,
#                        'WARNS' => <number>,
#                        'FIPS' => <number>,
#                        'NOTICES' => <number>,
#                        'QUELLED' => <number>,
#                        'SKIPPED' => <number>
#                    }
#                };
my %journal_data;

# Information about categories.
# %categories: CAT_ID => {
#                  'DESCRIPTION' => <text>,          # Short descriptive information shown below category title
#                  'ELEM_COL_NAME' => <name>         # Overridden title of the 'Element' column for the category table
#              };
my %categories = (
	# Problems caused by test suite itself or some environment
	'Journal format error' => {
		'DESCRIPTION' => 'Journal file seems to be corrupted.',
		'ELEM_COL_NAME' => 'Element'
	},
	'Journal syntax error' => {
		'DESCRIPTION' => 'A line in the journal file has incorrect syntax.',
		'ELEM_COL_NAME' => 'Element'
	},
	'Test case controller error' => {
		'DESCRIPTION' => 'Error reported by Test Case Controller.',
		'ELEM_COL_NAME' => 'Element'
	},
	'Test suite error' => {
		'DESCRIPTION' => 'Error reported by Test Suite itself.',
		'ELEM_COL_NAME' => 'Element'
	},
	'Empty test case' => {
		'DESCRIPTION' => 'Empty test case encountered (no test points).',
		'ELEM_COL_NAME' => 'Element'
	},
	'Test suite run aborted' => {
		'DESCRIPTION' => 'Test suite was aborted during execution by the user.',
		'ELEM_COL_NAME' => 'Element'
	},
	'Test point restarted' => {
		'DESCRIPTION' => 'Test Point Start entry appeared again for the same test point.',
		'ELEM_COL_NAME' => 'Element'
	},
	'Unrecognized result code' => {
		'DESCRIPTION' => 'Test returned a result code not recognized by the Moblin AppChecker.',
		'ELEM_COL_NAME' => 'Element'
	},
	'Wrong archive format' => {
		'DESCRIPTION' => 'There are problems in the archive format.',
		'ELEM_COL_NAME' => 'Element'
	},
	# Problems found in the files tested
	# moblin-appchk
	'ELF header check failed' => {
		'DESCRIPTION' => 'Some ELF header fields differ from what Moblin declares.',
		'ELEM_COL_NAME' => 'Element'
	},
	'Dynamic segment check failed' => {
		'DESCRIPTION' => 'Problems encountered while checking PT_DYNAMIC segment (dynamic linking information).',
		'ELEM_COL_NAME' => 'Element'
	},
	'Incorrect program loader' => {
		'DESCRIPTION' => 'The application uses program loader different from what Moblin declares. Try compiling your application with <b>lsbcc</b> compiler from LSB SDK package.',
		'ELEM_COL_NAME' => 'Loader'
	},
	'NOTE section check failed' => {
		'DESCRIPTION' => 'Problems encountered while checking PT_NOTE segment (auxiliary information).',
		'ELEM_COL_NAME' => 'Element'
	},
	'TLS section check failed' => {
		'DESCRIPTION' => 'Problems encountered while checking PT_TLS segment (thread-local storage).',
		'ELEM_COL_NAME' => 'Element'
	},
	'GNU_STACK header check failed' => {
		'DESCRIPTION' => 'Problems encountered while checking PT_GNU_STACK segment.',
		'ELEM_COL_NAME' => 'Element'
	},
	'ELF section check failed' => {
		'DESCRIPTION' => 'Problems encountered while checking ELF sections.',
		'ELEM_COL_NAME' => 'Element'
	},
	'Non-Moblin ELF section' => {
		'DESCRIPTION' => 'ELF sections found that are not present in Moblin.',
		'ELEM_COL_NAME' => 'Element'
	},
	'Non-Moblin library used' => {
		'DESCRIPTION' => 'The application uses a library that is not present in Moblin.',
		'ELEM_COL_NAME' => 'Library'
	},
	'Non-Moblin interface used' => {
		'DESCRIPTION' => 'The application uses an interface that is not present in Moblin.',
		'ELEM_COL_NAME' => 'Interface'
	},
	'I/O problem' => {
		'DESCRIPTION' => 'Problems encountered while trying to read file contents.',
		'ELEM_COL_NAME' => 'Element'
	},
	'Unexpected interface version' => {
		'DESCRIPTION' => 'A versioned symbol is used, but the version is not Moblin-compliant.',
		'ELEM_COL_NAME' => 'Interface'
	},
	'Deprecated interface used' => {
		'DESCRIPTION' => 'Deprecated interface are not recommended to use, because they will soon be withdrawn from Moblin.',
		'ELEM_COL_NAME' => 'Interface'
	},
	'ioctl() encountered' => {
		'DESCRIPTION' => '<b>ioctl()</b> interface is used to control stream device directly. Static application analysis is unable to check whether this interface is used according to Moblin or not, so you need to check it yourself and provide explanatory comments.',
		'ELEM_COL_NAME' => 'Interface'
	},
	'dlopen() encountered' => {
		'DESCRIPTION' => '<b>dlopen()</b> interface is used to dynamically load shared objects at run-time. Static application analysis is unable to check whether this interface is used according to Moblin or not, so you need to check it yourself and provide explanatory comments.',
		'ELEM_COL_NAME' => 'Interface'
	},
	# Uncategorized
	'Other' => {
		'DESCRIPTION' => 'All other problems not included in the categories given above.',
		'ELEM_COL_NAME' => 'Element'
	}
);

# Information about subcategories.
# %subcategories: SUBCAT_ID => {
#                     'CATEGORY' => <CAT-ID>,           # Shown in Category column (unique category ID)
#                     'ELEM_LINK' => <http-link>,       # This link + element name == full link to Navigator
#                     'MESSAGE' => <msg>                # Message shown prior to full details
#                 };
my %subcategories = (
	# Problems caused by test suite itself or some environment
	'J_FORMAT_ERROR' => {
		'CATEGORY' => 'Journal format error',
		'ELEM_LINK' => undef,
		'MESSAGE' => 'The line is not of TETJ format:'
	},
	'J_SYNTAX_ERROR' => {
		'CATEGORY' => 'Journal syntax error',
		'ELEM_LINK' => undef,
		'MESSAGE' => 'The line format does not match its code:'
	},
	'J_TCC_MIDDLE' => {
		'CATEGORY' => 'Journal format error',
		'ELEM_LINK' => undef,
		'MESSAGE' => 'TCC Start entry appears in the middle of the journal:'
	},
	'J_TC_NOT_STARTED' => {
		'CATEGORY' => 'Journal format error',
		'ELEM_LINK' => undef,
		'MESSAGE' => 'Test Case End entry found when no test case seems to be running.'
	},
	'J_TC_NOT_FINISHED' => {
		'CATEGORY' => 'Journal format error',
		'ELEM_LINK' => undef,
		'MESSAGE' => 'Test case starts before the previous one has finished:'
	},
	'TS_CONTROLLER_MSG' => {
		'CATEGORY' => 'Test case controller error',
		'ELEM_LINK' => undef,
		'MESSAGE' => 'Test case controller reports a problem:'
	},
	'TS_ERROR' => {
		'CATEGORY' => 'Test suite error',
		'ELEM_LINK' => undef,
		'MESSAGE' => 'Error reported by the test suite:'
	},
	'R_EMPTY_TEST_CASE' => {
		'CATEGORY' => 'Empty test case',
		'ELEM_LINK' => undef,
		'MESSAGE' => 'This test case contains no test points.'
	},
	'TS_ABORTED' => {
		'CATEGORY' => 'Test suite run aborted',
		'ELEM_LINK' => undef,
		'MESSAGE' => 'Test suite execution interrupted by the user.'
	},
	'J_TP_RESTARTED' => {
		'CATEGORY' => 'Test point restarted',
		'ELEM_LINK' => undef,
		'MESSAGE' => 'Test was restarted in the middle.'
	},
	'J_TP_NOT_STARTED' => {
		'CATEGORY' => 'Journal format error',
		'ELEM_LINK' => undef,
		'MESSAGE' => 'Test result entry found but the test point had not started.'
	},
	'J_TP_NOT_FINISHED' => {
		'CATEGORY' => 'Journal format error',
		'ELEM_LINK' => undef,
		'MESSAGE' => 'Test point starts before the previous one has finished:'
	},
	'J_TP_CHANGED' => {
		'CATEGORY' => 'Journal format error',
		'ELEM_LINK' => undef,
		'MESSAGE' => 'Test point number at the end is different from the starting one:'
	},
	'R_UNKNOWN_RESULT' => {
		'CATEGORY' => 'Unrecognized result code',
		'ELEM_LINK' => undef,
		'MESSAGE' => 'Unrecognized test result code.'
	},
	'R_FAILED_UNPACK' => {
		'CATEGORY' => 'Wrong archive format',
		'ELEM_LINK' => undef,
		'MESSAGE' => 'The archive could not be unpacked:'
	},
	'R_NON_MOBLIN_ARCHIVE' => {
		'CATEGORY' => 'Wrong archive format',
		'ELEM_LINK' => undef,
		'MESSAGE' => 'The archive is not Moblin compliant:'
	},
	# Problems found in the files tested
	# moblin-appchk
	'R_ELF_HDR_FAILED' => {
		'CATEGORY' => 'ELF header check failed',
		'ELEM_LINK' => undef,
		'MESSAGE' => 'Messages from the test:'
	},
	'R_LOAD_HDR_FAILED' => {
		'CATEGORY' => 'ELF header check failed',
		'ELEM_LINK' => undef,
		'MESSAGE' => 'Messages from the test:'
	},
	'R_DYN_FAILED' => {
		'CATEGORY' => 'Dynamic segment check failed',
		'ELEM_LINK' => undef,
		'MESSAGE' => 'Messages from the test:'
	},
	'R_WRONG_LOADER' => {
		'CATEGORY' => 'Incorrect program loader',
		'ELEM_LINK' => undef,
		'MESSAGE' => 'For Moblin applications the program loader should be something like <b>/lib/ld-lsb.so.3</b>.<br />Messages from the test:'
	},
	'R_NOTE_FAILED' => {
		'CATEGORY' => 'NOTE section check failed',
		'ELEM_LINK' => undef,
		'MESSAGE' => 'Messages from the test:'
	},
	'R_TLS_FAILED' => {
		'CATEGORY' => 'TLS section check failed',
		'ELEM_LINK' => undef,
		'MESSAGE' => 'Messages from the test:'
	},
	'R_GNU_STACK_FAILED' => {
		'CATEGORY' => 'GNU_STACK header check failed',
		'ELEM_LINK' => undef,
		'MESSAGE' => 'Messages from the test:'
	},
	'R_ELF_SECTION_FAILED' => {
		'CATEGORY' => 'ELF section check failed',
		'ELEM_LINK' => undef,
		'MESSAGE' => 'Messages from the test:'
	},
	'R_ELF_SECTION_NON_MOBLIN' => {
		'CATEGORY' => 'Non-Moblin ELF section',
		'ELEM_LINK' => undef,
		'MESSAGE' => 'Messages from the test:'
	},
	'R_LIBRARY_NON_MOBLIN' => {
		'CATEGORY' => 'Non-Moblin library used',
		'ELEM_LINK' => 'http://www.linuxfoundation.org/navigator/browse/rawlib.php?cmd=list-byrunname&amp;RLrunname=',
		'MESSAGE' => 'Messages from the test:'
	},
	'R_INTERFACE_NON_MOBLIN' => {
		'CATEGORY' => 'Non-Moblin interface used',
		'ELEM_LINK' => 'http://www.linuxfoundation.org/navigator/browse/int_single.php?cmd=list-by-name&amp;Iname=',
		'MESSAGE' => 'Messages from the test:'
	},
	'R_IO_FAILED' => {
		'CATEGORY' => 'I/O problem',
		'ELEM_LINK' => undef,
		'MESSAGE' => 'Messages from the test:'
	},
	'R_INTERFACE_VERSION_FAILED' => {
		'CATEGORY' => 'Unexpected interface version',
		'ELEM_LINK' => 'http://www.linuxfoundation.org/navigator/browse/int_single.php?cmd=list-by-name&amp;Iname=',
		'MESSAGE' => 'Messages from the test:'
	},
	'R_INTERFACE_DEPRECATED' => {
		'CATEGORY' => 'Deprecated interface used',
		'ELEM_LINK' => 'http://www.linuxfoundation.org/navigator/browse/int_single.php?cmd=list-by-name&amp;Iname=',
		'MESSAGE' => 'Messages from the test:'
	},
	'R_IOCTL_FIP' => {
		'CATEGORY' => 'ioctl() encountered',
		'ELEM_LINK' => 'http://www.linuxfoundation.org/navigator/browse/int_single.php?cmd=list-by-name&amp;Iname=',
		'MESSAGE' => 'Messages from the test:'
	},
	'R_DLOPEN_FIP' => {
		'CATEGORY' => 'dlopen() encountered',
		'ELEM_LINK' => 'http://www.linuxfoundation.org/navigator/browse/int_single.php?cmd=list-by-name&amp;Iname=',
		'MESSAGE' => 'Messages from the test:'
	},
	# Uncategorized
	'R_OTHER' => {
		'CATEGORY' => 'Other',
		'ELEM_LINK' => undef,
		'MESSAGE' => 'Messages from the test:'
	}
);

# Sorted list of categories
my @categories_sort = (
	# Problems caused by test suite itself or some environment
	'Journal format error',
	'Journal syntax error',
	'Test case controller error',
	'Test suite error',
	'Empty test case',
	'Test suite run aborted',
	'Test point restarted',
	'Unrecognized result code',
	'Wrong archive format',
	# Problems found in the files tested
	# moblin-appchk
	'ELF header check failed',
	'Dynamic segment check failed',
	'Incorrect program loader',
	'NOTE section check failed',
	'TLS section check failed',
	'GNU_STACK header check failed',
	'ELF section check failed',
	'Non-Moblin ELF section',
	'Non-Moblin library used',
	'Non-Moblin interface used',
	'I/O problem',
	'Unexpected interface version',
	'Deprecated interface used',
	'ioctl() encountered',
	'dlopen() encountered',
	# Uncategorized
	'Other'
);

# Reversed hash: problems list for each category
# %cat_problems: <CAT-ID> => {
#                    'FAILS' => <number>            # Total number of fails in this category
#                    'PROBLEMS' => [                # List of problems
#                        {
#                            'JOURNAL_TITLE' => <journal_title>     # Reference to journal info
#                            'PROBLEM_IDX' => <idx>                 # Index of the specific problem
#                        },
#                        ...
#                    ]
#                }
my %cat_problems = ();

# Reversed hash for retrieving journal title from HTML anchor.
my %journal_title_from_anchor;

# Map of TET result codes to human-readable severity names.
my %result_code_map = (
	0, "Passed",
	1, "Failed",
	2, "Unresolved",
	3, "Unused",
	4, "Unsupported",
	5, "Untested",
	6, "Uninitiated",
	7, "Unreported",
	65, "Time Expired",
	101, "Warning",
	102, "FIP",
	103, "Unimplemented",
	104, "Unapproved"
);

# Global statistics
my $tests_cancelled = 0;    # Number of cancelled tests

################################################################################
# Functions

# Prints usage info and exits.
sub usage() {
	print "Usage: $0 <test_journals_dir> [-o output-file.tgz] [-n app-name]\n";
	exit(1);
}

# Parses command line options passed to the script.
sub parse_cmd_line() {
	my $argc = scalar(@ARGV);
	if ($argc < 1) {
		usage();
	}

	for (my $i = 0; $i < $argc; ++$i) {
		if ($ARGV[$i] eq '-o') {
			usage() if ($i == $argc - 1);
			++$i;
			$result_archive = $ARGV[$i];
		}
		elsif ($ARGV[$i] eq '-n') {
			usage() if ($i == $argc - 1);
			++$i;
			$application_name = $ARGV[$i];
		}
		else {
			$result_path = abs_path($ARGV[$i]);
			$result_path =~ s!/+$!!;
			$result_path =~ m!^(.*)/([^/]+)$!;
			$result_path = ($1 or '/');
			$result_dir = $2;
		}
	}
	unless ($result_archive) {
		$result_archive = "$result_dir.tgz";
	}
}

# Reads the whole file contents.
sub read_file($) {
	my ($file) = @_;
	my $res = '';
	if (open(FILE, $file)) {
		read(FILE, $res, -s $file);
		close(FILE);
	}
	else {
		print "Couldn't open file '$file' for reading: $!";
	}
	return $res;
}

#Reads information about host system
sub read_host_info() {
	$moblin_version = (read_file("$result_path/$result_dir/test_logs/moblin_version") or '4.0');
	$moblin_version =~ s/[\r\n]//g;

	$test_comments = read_file("$result_path/$result_dir/test_logs/test_comments");

	$host_machine  = `uname -m`;
	$host_machine  =~ s/[\r\n]//g;

	$host_name     = `uname -n`;
	$host_name     =~ s/[\r\n]//g;

	$test_arch     = (read_file("$result_path/$result_dir/test_logs/host_platform") or $host_machine);
	$test_arch     =~ s/[\r\n]//g;

	$host_os = `lsb_release -s -d`;
	$host_os =~ s/\"(.*)\"/$1/;	# remove quotes
	$host_os =~ s/[\r\n]//g;

	$host_kernel = `uname -r -v`;
	$host_kernel =~ s/[\r\n]//g;
}

# Reads all journals and generates report
sub generate_report() {
	print "Generating HTML report\n";

	# Go to the output directory
	chdir("$result_path/$result_dir");

	# Parse each test journal
	opendir(DIR, 'test_journals') or die "Can't open directory $result_path/$result_dir/test_journals: $!\n";
	my @files = readdir(DIR);
	closedir(DIR);
	foreach my $file (@files) {
		if ($file =~ m/^(.*\.INFO)$/) {
			parse_journal("test_journals/$file");
		}
	}
	
	# If application name was not specified, use all journal titles.
	if (!$application_name) {
		$application_name = join(", ", @journal_titles);
	}

	# Sum up statistics
	my $total_stats = sum_stats();

	# Give the final verdict
	my $overall_verdict;
	if (($total_stats->{'FAILS'} > 0) or ($tests_cancelled > 0)) {
		$overall_verdict = 'Failed';
	}
	elsif (($total_stats->{'WARNS'} > 0) or ($total_stats->{'FIPS'} > 0)) {
		$overall_verdict = 'Warnings';
	}
	elsif (($total_stats->{'TOTAL'} == 0)) {
		$overall_verdict = 'No result';
	}
	else {
		$overall_verdict = 'Success';
	}

	# Write INFO_MOBLIN file
	my $INFO_filename = "$result_path/$result_dir/INFO_MOBLIN";
	if (open(INFO, ">$INFO_filename")) {
		print INFO "Verdict: $overall_verdict\n";
		close(INFO);
	}
	else {
		print "Couldn't open file '$INFO_filename' for writing: $!";
	}

	compile_report();

	# Create TAR.GZ
	my $full_result_archive = "$result_path/$result_dir/$result_archive";
	if (-e "$full_result_archive") {
		unlink("$full_result_archive");
		if (-f $full_result_archive) {
			die "Failed to remove old archive.\n";
		}
	}
	chdir($result_path);
	my $result_archive_esc = escape_path($result_archive);
	my $result_dir_esc = escape_path($result_dir);
	system("tar czf $result_archive_esc $result_dir_esc && mv $result_archive_esc $result_dir_esc");
}

# TREE: {<node_name> => {'type' => (file|dir|pkg), 'children' => TREE, 'journal_anchor' => <anchor>}};
my %journal_tree;

# Adds element to the branch of journal tree.
sub add_elem_to_subtree($$$$);  # Prototype for recursive function
sub add_elem_to_subtree($$$$) {
	my ($subtree, $elem_name, $f_type, $j_anchor) = @_;
	$elem_name =~ s/^\/+//;
	my $idx = index($elem_name, '/');
	if ($idx > 0) {
		my $d_name = substr($elem_name, 0, $idx);
		$elem_name = substr($elem_name, $idx + 1);
		if (!$subtree->{'children'}->{$d_name}->{'type'}) {
			$subtree->{'children'}->{$d_name}->{'type'} = 'dir';
		}
		add_elem_to_subtree($subtree->{'children'}->{$d_name}, $elem_name, $f_type, $j_anchor);
	}
	else {
		$subtree->{'children'}->{$elem_name}->{'type'} = $f_type;
		$subtree->{'children'}->{$elem_name}->{'journal_anchor'} = $j_anchor;
	}
}

# Adds element to the journal tree.
sub add_elem_to_tree($$$) {
	my ($title, $anchor, $file) = @_;
	my $type = ((substr($title, -1) eq '|') ? 'pkg' : 'file');
	my @tmp = split(/\|/, $title, 2);
	if ($tmp[1]) {
		$tmp[1] =~ s/\|//g;
		$title = $tmp[0].'|'.$tmp[1];
	}
	else {
		$title = $tmp[0];
	}

	if ($title =~ m/^[^|]+$/) {
   		$journal_tree{$title}->{'type'} = $type;
   		$journal_tree{$title}->{'journal_anchor'} = $anchor;
	}
	elsif ($title =~ m/^([^|]+)\|(.+)$/) {
		my ($tree_root, $elem_name) = ($1, $2);
   		if (!$journal_tree{$tree_root}) {
			if ($tree_root =~ m/^PKG: /) {
				$journal_tree{$tree_root}->{'type'} = 'pkg';
			}
			else {
				$journal_tree{$tree_root}->{'type'} = 'dir';
			}
   		}
   		add_elem_to_subtree($journal_tree{$tree_root}, $elem_name, $type, $anchor);
	}
	else {
		print "Internal error: wrong format for TITLE ($title).\nFile: $file\n";
	}
}

# Map from journal status code to statistics category
my %stats_map = (
	'passed' => 'PASSES',
	'failed' => 'FAILS',
	'unresolved' => 'FAILS',
	'unused' => 'SKIPPED',
	'unsupported' => 'SKIPPED',
	'untested' => 'NOTICES',
	'uninitiated' => 'FAILS',
	'unreported' => 'FAILS',
	'time expired' => 'FAILS',
	'warning' => 'WARNS',
	'fip' => 'FIPS',
	'unimplemented' => 'SKIPPED',
	'unapproved' => 'SKIPPED',
	'test suite error' => 'FAILS',
	'notice' => 'NOTICES'
);

# Adds a new problem entry into parsed journal info
sub add_journal_entry($$$$$$$) {
	my ($journal_title_data, $severity, $test_case, $subcategory, $element, $comment, $details) = @_;
	my $lc_severity = lc($severity);
	my $stat_name = $stats_map{$lc_severity};
	if ($stat_name) {
		++$journal_title_data->{'STATISTICS'}->{$stat_name};
		if (($stat_name eq 'SKIPPED') or ($stat_name eq 'PASSES')) {
			$severity = '';     # Ignore this entry
		}
	}
	else {
		$severity = "Unrecognized ($severity)";
	}
	if ($severity) {
		push @{$journal_title_data->{'PROBLEMS_LIST'}}, {
			'SEVERITY' => $severity,
			'TEST_CASE' => $test_case,
			'SUBCATEGORY' => $subcategory,
			'ELEMENT' => $element,
			'COMMENT' => $comment,
			'DETAILS' => $details
		};
	}
}

# Analyzes the test messages and returns an array of problem entries
sub guess_problem_subcategory($$) {
	my ($test_messages, $test_tool) = @_;
	if (!$test_messages) {
		return [{'SUBCATEGORY' => 'R_OTHER', 'ELEMENT' => '', 'MESSAGE' => ''}];
	}
	my @msgs = split(/\n/, $test_messages);
	my $res = [];
	my $current = {'SUBCATEGORY' => '', 'ELEMENT' => '', 'MESSAGE' => ''};
	foreach my $msg (@msgs) {
		my $tmp = guess_problem_subcategory_line($msg, $test_tool);
		if (($tmp->{'SUBCATEGORY'} eq 'R_LIBRARY_NON_MOBLIN') or
			($tmp->{'SUBCATEGORY'} eq 'R_RPM_DEPS_FAILED') or
			($tmp->{'SUBCATEGORY'} eq 'R_RPM_DEPS_FAILED_LINK') or
			($tmp->{'SUBCATEGORY'} eq 'R_DYN_FAILED')
			) {
			if ($current->{'SUBCATEGORY'}) {
				push @$res, $current;
				$current = {'SUBCATEGORY' => '', 'ELEMENT' => '', 'MESSAGE' => ''};
			}
			push @$res, {'SUBCATEGORY' => $tmp->{'SUBCATEGORY'}, 'ELEMENT' => $tmp->{'ELEMENT'}, 'MESSAGE' => $msg};
		}
		else {
			$current->{'MESSAGE'} .= ($current->{'MESSAGE'} ? "\n" : '').$msg;
			if (!$current->{'SUBCATEGORY'}) {
				$current->{'SUBCATEGORY'} = $tmp->{'SUBCATEGORY'};
				$current->{'ELEMENT'} = $tmp->{'ELEMENT'};
			}
			else {
				if (!$current->{'ELEMENT'}) {
					$current->{'ELEMENT'} = $tmp->{'ELEMENT'};
				}
				elsif ($current->{'ELEMENT'} ne $tmp->{'ELEMENT'}) {
					$current->{'ELEMENT'} = '&hellip;';
				}
			}
		}
	}
	if ($current->{'SUBCATEGORY'}) {
		push @$res, $current;
	}
	return $res;
}

# Analyzes the test message line and determines its subcategory and significant element
sub guess_problem_subcategory_line($$) {
	my ($msg_line, $test_tool) = @_;
	if (!$msg_line) {
		return {'SUBCATEGORY' => 'R_OTHER', 'ELEMENT' => ''};
	}
	my $res;
	if ($msg_line =~ m/Failed to unpack the archive\./i) {
		$res = {'SUBCATEGORY' => 'R_FAILED_UNPACK', 'ELEMENT' => ''};
	}
	elsif ($msg_line =~ m/No contents found in the DEB archive\./i) {
		$res = {'SUBCATEGORY' => 'R_FAILED_UNPACK', 'ELEMENT' => ''};
	}
	elsif ($msg_line =~ m/.* archives are not standardized by the Moblin\./i) {
		$res = {'SUBCATEGORY' => 'R_NON_MOBLIN_ARCHIVE', 'ELEMENT' => ''};
	}
	# moblin-appchk
	elsif ($msg_line =~ m/compareElfhdr: (.*) isn\'t expected value /i) {
		$res = {'SUBCATEGORY' => 'R_ELF_HDR_FAILED', 'ELEMENT' => $1};
	}
	elsif ($msg_line =~ m/Section (.*) does not fit in Segment/i) {
		$res = {'SUBCATEGORY' => 'R_LOAD_HDR_FAILED', 'ELEMENT' => "Section $1"};
	}
	elsif ($msg_line =~ m/Section (.*) flags .* does not correspond to Segment flags /i) {
		$res = {'SUBCATEGORY' => 'R_LOAD_HDR_FAILED', 'ELEMENT' => "Section $1"};
	}
	elsif ($msg_line =~ m/Dynamic section (.*) does not match Segment /i) {
		$res = {'SUBCATEGORY' => 'R_DYN_FAILED', 'ELEMENT' => "Section $1"};
	}
	elsif ($msg_line =~ m/Non-dynamic section (.*) found in PT_DYNAMIC segment/i) {
		$res = {'SUBCATEGORY' => 'R_DYN_FAILED', 'ELEMENT' => "Section $1"};
	}
	elsif ($msg_line =~ m/Incorrect program interpreter: (.*)/i) {
		$res = {'SUBCATEGORY' => 'R_WRONG_LOADER', 'ELEMENT' => $1};
	}
	elsif ($msg_line =~ m/NOTE (section .*) does not match Segment /i) {
		$res = {'SUBCATEGORY' => 'R_NOTE_FAILED', 'ELEMENT' => $1};
	}
	elsif ($msg_line =~ m/Section (.*) in PT_NOTE segment is not type SHT_NOTE/i) {
		$res = {'SUBCATEGORY' => 'R_NOTE_FAILED', 'ELEMENT' => "Section $1"};
	}
	elsif ($msg_line =~ m/PT_SHLIB is not permitted!/i) {
		$res = {'SUBCATEGORY' => 'R_OTHER', 'ELEMENT' => ''};
	}
	elsif ($msg_line =~ m/PHDR offset doesn\'t match expected /i) {
		$res = {'SUBCATEGORY' => 'R_OTHER', 'ELEMENT' => ''};
	}
	elsif ($msg_line =~ m/TLS (section .*) does not match Segment /i) {
		$res = {'SUBCATEGORY' => 'R_TLS_FAILED', 'ELEMENT' => $1};
	}
	elsif ($msg_line =~ m/Section (.*) in PT_TLS segment is not type SHF_TLS/i) {
		$res = {'SUBCATEGORY' => 'R_TLS_FAILED', 'ELEMENT' => "Section $1"};
	}
	elsif ($msg_line =~ m/Section in PT_GNU_EH_FRAME segment is not ".eh_frame_hdr"/i) {
		$res = {'SUBCATEGORY' => 'R_OTHER', 'ELEMENT' => ''};
	}
	elsif ($msg_line =~ m/(offset|vaddr|paddr|filesz|memsz) not 0/i) {
		$res = {'SUBCATEGORY' => 'R_GNU_STACK_FAILED', 'ELEMENT' => $1};
	}
	elsif ($msg_line =~ m/Unexpected flags /i) {
		$res = {'SUBCATEGORY' => 'R_GNU_STACK_FAILED', 'ELEMENT' => 'flags'};
	}
	elsif ($msg_line =~ m/IA_64_UNWIND not checked!/i) {
		$res = {'SUBCATEGORY' => 'R_OTHER', 'ELEMENT' => ''};
	}
	elsif ($msg_line =~ m/Section (.*): .* is wrong\. /i) {
		$res = {'SUBCATEGORY' => 'R_ELF_SECTION_FAILED', 'ELEMENT' => $1};
	}
	elsif ($msg_line =~ m/Section (.*) not checked/i) {
		$res = {'SUBCATEGORY' => 'R_ELF_SECTION_FAILED', 'ELEMENT' => $1};
	}
	elsif ($msg_line =~ m/section (.*) is not in the Moblin/i) {
		$res = {'SUBCATEGORY' => 'R_ELF_SECTION_NON_MOBLIN', 'ELEMENT' => $1};
	}
	elsif ($msg_line =~ m/ignoring (.*): Not in the Moblin but not loaded/i) {
		$res = {'SUBCATEGORY' => 'R_ELF_SECTION_NON_MOBLIN', 'ELEMENT' => $1};
	}
	elsif ($msg_line =~ m/DT_NEEDED: (.*) is used, but not part of the Moblin/i) {
		$res = {'SUBCATEGORY' => 'R_LIBRARY_NON_MOBLIN', 'ELEMENT' => $1};
	}
	elsif ($msg_line =~ m/Dynamic Tag .* unknown/i) {
		$res = {'SUBCATEGORY' => 'R_OTHER', 'ELEMENT' => 'Dynamic Tag'};
	}
	elsif ($msg_line =~ m/Could not open (.*)/i) {
		$res = {'SUBCATEGORY' => 'R_IO_FAILED', 'ELEMENT' => $1};
	}
	elsif ($msg_line =~ m/Could not stat file/i) {
		$res = {'SUBCATEGORY' => 'R_IO_FAILED', 'ELEMENT' => ''};
	}
	elsif ($msg_line =~ m/Symbol (.*) is used, but is not included in /i) {
		$res = {'SUBCATEGORY' => 'R_INTERFACE_NON_MOBLIN', 'ELEMENT' => $1};
	}
	elsif ($msg_line =~ m/Symbol (.*) has version .* expecting /i) {
		$res = {'SUBCATEGORY' => 'R_INTERFACE_VERSION_FAILED', 'ELEMENT' => $1};
	}
	elsif ($msg_line =~ m/Symbol (.*) has been deprecated/i) {
		$res = {'SUBCATEGORY' => 'R_INTERFACE_DEPRECATED', 'ELEMENT' => $1};
	}
	elsif ($msg_line =~ m/Unable to determine if parameters to ioctl\(\) are used in accordance with the Moblin/i) {
		$res = {'SUBCATEGORY' => 'R_IOCTL_FIP', 'ELEMENT' => 'ioctl'};
	}
	elsif ($msg_line =~ m/Unable to automatically check libraries accessed via dlopen\(\)/i) {
		$res = {'SUBCATEGORY' => 'R_DLOPEN_FIP', 'ELEMENT' => 'dlopen'};
	}
	# Uncategorized
	if (!$res) {
		$res = {'SUBCATEGORY' => 'R_OTHER', 'ELEMENT' => ''};
	}

	return $res;
}

# Reads and parses specific TET journal specified by the INFO file.
sub parse_journal($) {
	my ($INFO_filename) = @_;
	my $journal_file;
	my $journal_title;
	my $journal_anchor;

	# Parse INFO file
	if (open(FILE, $INFO_filename)) {
		while (my $line = <FILE>) {
			if ($line =~ m/^\s*TITLE:\s+(.*)$/) {
				$journal_title = $1;
			}
			elsif ($line =~ m/^\s*ANCHOR:\s+(.*)$/) {
				$journal_anchor = $1;
			}
			elsif ($line =~ m/^\s*JOURNAL_FILENAME:\s+(.*)$/) {
				$journal_file = "test_journals/$1";
			}
		}
		close FILE;
	}
	else {
		print "Couldn't open file '$INFO_filename' for reading: $!";
	}
	if (!$journal_file) {
		$INFO_filename =~ m/^(.*)\.INFO$/;
		$journal_file = $1;
	}
	if (!$journal_anchor) {
		$journal_anchor = $journal_file;
		$journal_anchor =~ s/.*\///;    # Cut path
		$journal_anchor =~ s/\..*$//;   # Cut extension
	}
	if (!$journal_title) {
		$journal_title = $journal_anchor;
	}

	add_elem_to_tree($journal_title, $journal_anchor, $journal_file);

	# Store the journal title and anchor.
	push @journal_titles, $journal_title;
	$journal_title_from_anchor{$journal_anchor} = $journal_title;

	$journal_data{$journal_title} = {
		'JOURNAL_FILE' => $journal_file,
		'TEST_TOOL' => '&lt;Unknown&gt;',
		'START_TIME' => '&lt;Unknown&gt;',
		'PROBLEMS_LIST' => [],
		'IS_RESULT' => 1,
		'STATISTICS' => {
			'PASSES' => 0,
			'FAILS' => 0,
			'WARNS' => 0,
			'FIPS' => 0,
			'NOTICES' => 0,
			'QUELLED' => 0,
			'SKIPPED' => 0
		}
	};

	# Open the journal file.
	print "Reading journal: $journal_file\n";
	if (!open(JOURNAL, $journal_file)) {
		# The journal file could not be open; mention this in the log
		# (so that we know that something was missing) and return.

		my $journal_title_disp = $journal_title;
		$journal_title_disp =~ s/^\|//;
		$journal_title_disp =~ s/\|$//;
		$journal_title_disp =~ s/\|/ =&gt; /g;

		++$tests_cancelled;

		print "Failed to read journal: $journal_file.\n$journal_title_disp failed to run, interrupted or crashed during the execution.\n";
		$journal_data{$journal_title}->{'IS_RESULT'} = undef;
		return;
	}

	# This flag is set to nonzero if we are inside a test case.
	my $inside_test = 0;

	# Variables for storing information about test cases
	my $test_name;
	my $test_number;
	my $test_point;
	my $test_points_in_this_case;
	my $test_messages;
	my $line_number = 0;
	my $journal_empty = 1;

	# List of problems in the current journal
	my $journal_title_data = $journal_data{$journal_title};

	# Read the journal line by line.
	while (my $line = <JOURNAL>) {
		++$line_number;

		# Each journal line consists of three parts, separated by
		# a vertical bar:
		#
		# 1. A numeric line code, determining the meaning of the line.
		# 2. Additional data, format depends on line code.
		# 3. Additional text (message for the user, or some comment).
		unless ($line =~ /^(\d*)\|([^|]*)\|(.*)$/) {
			# Lines that do not conform to this format are reported
			# to the user and skipped.
			add_journal_entry(
				$journal_title_data,
				'Warning',
				'',
				'J_FORMAT_ERROR',
				"$journal_file:$line_number",
				'',
				$line
			);
			next;
		}
		my $line_code = $1;
		my $line_data = $2;
		my $line_text = $3;

		# Determine the type of the line and process it appropriately.
		if ($line_code == 0) {
			# Code 0 -- Test Case Controller Start.
			# The line structure is:
			# 0|version time date|who
			unless ($line_data =~ /^(\S*)\s*(\S*)\s(\S*)$/) {
				add_journal_entry(
					$journal_title_data,
					'Warning',
					'',
					'J_SYNTAX_ERROR',
					"$journal_file:$line_number",
					'',
					$line
				);
				next;
			}

			# This line should be one of the first;
			# if it comes in the middle of normal results, it is strange.
			if ($inside_test) {
				add_journal_entry(
					$journal_title_data,
					'Warning',
					'',
					'J_TCC_MIDDLE',
					"$journal_file:$line_number",
					'',
					"$line_data\n$line_text"
				);
			}
			else {
				# my $test_engine_version = $1;
				my $test_start_time = $2;
				my $test_start_date = $3;
				if ($test_start_date =~ m/^(\d{4})(\d{2})(\d{2})$/) {
					$test_start_date = "$3.$2.$1";
				}
				$journal_title_data->{'START_TIME'} = "$test_start_date $test_start_time";

				$line_text =~ s/^.*Command line:\s*//;
				if ($line_text =~ m!^\S*/([^\s/]+)\s?!) {
					$line_text = $1;
				}
				if ($line_text eq 'moblinappchk') {
					$journal_title_data->{'TEST_TOOL'} = 'moblin-appchk';
				}
				elsif ($line_text eq '<none>') {
					$journal_title_data->{'TEST_TOOL'} = 'None';
				}
				else {
					$journal_title_data->{'TEST_TOOL'} = "Unknown ($line_text)";
				}
			}
		}
		elsif ($line_code == 5) {
			# Code 5 -- Local System Information.
			# Ignore this data - it's the same for all journals.
		}
		elsif ($line_code == 10) {
			# Code 10 -- Test Case Start.
			# The line structure is:
			# 10|activitynum testname hh:mm:ss|list of invocable components
			unless ($line =~ /^\d*\|(\d*)\s*(.*)\s+\S+\|/) {
				add_journal_entry(
					$journal_title_data,
					'Warning',
					'',
					'J_SYNTAX_ERROR',
					"$journal_file:$line_number",
					'',
					$line
				);
				next;
			}

			if ($test_name) {
				# The previous test case was not properly finished;
				# issue a warning.
				add_journal_entry(
					$journal_title_data,
					'Warning',
					"$test_name:0",
					'J_TC_NOT_FINISHED',
					"$journal_file:$line_number",
					'',
					"Previous test case name: $test_name"
				);
			}

			# Store the test name and number.
			$test_number = $1;
			$test_name = $2;
			$test_name =~ s|^.*/([^/]+)$|$1|;
			$test_name =~ s|^(.*)-pass2$|$1|;
			$inside_test = 1;

			# Reset the test point number and counter.
			$test_point = 0;
			$test_points_in_this_case = 0;
		}
		elsif ($line_code == 15) {
			# Code 15 -- Test Case Manager Start.
			# Not interesting to us.
		}
		elsif ($line_code == 20) {
			# Code 20 -- Local/Remote System Configuration Start.
			# Not interesting to us.
		}
		elsif ($line_code == 30) {
			# Code 30 -- Configuration Variables.
			# Not interesting to us.
		}
		elsif ($line_code == 40) {
			# Code 40 -- Configuration End.
			# Not interesting to us.
		}
		elsif ($line_code == 50) {
			# Code 50 -- Test Case Controller Message.
			add_journal_entry(
				$journal_title_data,
				'Test Suite Error',
				"$test_name:$test_point",
				'TS_CONTROLLER_MSG',
				"$journal_file:$line_number",
				'',
				$line_text
			);
		}
		elsif ($line_code == 70) {
			# Code 70 -- Scenario Information.
			# Not interesting to us.
		}
		elsif ($line_code == 80) {
			# Code 80 -- Test Case End.
			unless ($test_name) {
				# Test case end found when no test case was started;
				# issue a warning.
				add_journal_entry(
					$journal_title_data,
					'Warning',
					'',
					'J_TC_NOT_STARTED',
					"$journal_file:$line_number",
					'',
					''
				);
				next;
			}

			# Warn if the test case seems to be empty.
			if ($test_points_in_this_case == 0) {
				add_journal_entry(
					$journal_title_data,
					'Warning',
					$test_name,
					'R_EMPTY_TEST_CASE',
					"$journal_file:$line_number",
					'',
					''
				);
			}

			# Clear the test name, number, and messages.
			$test_name = '';
			$test_number = 0;
			$test_messages = '';
			$test_points_in_this_case = 0;
		}
		elsif ($line_code == 90) {
			# Code 90 -- User Abort.
			add_journal_entry(
				$journal_title_data,
				'Test Suite Error',
				'',
				'TS_ABORTED',
				"$journal_file:$line_number",
				'',
				''
			);
		}
		elsif ($line_code == 100) {
			# Code 100 -- Captured Output.
			# Not interesting to us.
		}
		elsif ($line_code == 110) {
			# Code 110 -- Build Start.
			# Not interesting to us.
		}
		elsif ($line_code == 130) {
			# Code 130 -- Build End.
			# Not interesting to us.
		}
		elsif ($line_code == 200) {
			# Code 200 -- Test Purpose Start.
			# The structure of the line is as follows:
			# 200|activitynum pointnum hh:mm:ss|comment
			unless ($line =~ /^\d*\|(\d*)\s*(\d*)/) {
				add_journal_entry(
					$journal_title_data,
					'Warning',
					'',
					'J_SYNTAX_ERROR',
					"$journal_file:$line_number",
					'',
					$line
				);
				next;
			}

			# Check if the previous test point was properly finished.
			if ($test_point > 0) {
				if ($test_point == $2) {
					# Some test suites have a strange habit of re-starting
					# the same test point more than once.
					add_journal_entry(
						$journal_title_data,
						'Notice',
						"$test_name:$test_point",
						'J_TP_RESTARTED',
						"$journal_file:$line_number",
						'',
						''
					);
				}
				else {
					# Notify about a disappeared test.
					add_journal_entry(
						$journal_title_data,
						'Warning',
						"$test_name:$2",
						'J_TP_NOT_FINISHED',
						"$journal_file:$line_number",
						'',
						"Previous test point number: $test_point"
					);
				}
			}

			# Clear the messages.
			$test_messages = '';

			# Store the test point number.
			$test_point = $2;

			# Increment the number of test points.
			++$test_points_in_this_case;
		}
		elsif ($line_code == 220) {
			# Code 220 -- Test Purpose Result.
			# The structure of the line is as follows:
			# 220|activitynum pointnum result time|comment
			unless ($line =~ /^\d*\|(\d*)\s*(\d*)\s*(\d*)/) {
				add_journal_entry(
					$journal_title_data,
					'Warning',
					'',
					'J_SYNTAX_ERROR',
					"$journal_file:$line_number",
					'',
					$line
				);
				next;
			}

			# Print a warning if the test purpose seems to never been started.
			if ($test_point == 0) {
				add_journal_entry(
					$journal_title_data,
					'Warning',
					"$test_name:$test_point",
					'J_TP_NOT_STARTED',
					"$journal_file:$line_number",
					'',
					''
				);
			}

			# Print a warning if the test purpose number is different from
			# what we expected.
			elsif ($test_point != $2) {
				add_journal_entry(
					$journal_title_data,
					'Warning',
					"$test_name:$test_point",
					'J_TP_NOT_STARTED',
					"$journal_file:$line_number",
					'',
					"Started with TP $test_point; ended with TP $2."
				);
			}

			# Get the result code.
			my $result_code = $3;

			# Find the meaning of the result code (severity).
			my $severity = $result_code_map{$result_code};
			if ($severity) {
				my $cat_infos = guess_problem_subcategory($test_messages, $journal_title_data->{'TEST_TOOL'});
				foreach my $info (@$cat_infos) {
					add_journal_entry(
						$journal_title_data,
						$severity,
						"$test_name:$test_point",
						$info->{'SUBCATEGORY'},
						$info->{'ELEMENT'},
						'',
						$info->{'MESSAGE'}
					);
				}
			}
			else {
				add_journal_entry(
					$journal_title_data,
					'Failed',
					"$test_name:$test_point",
					'R_UNKNOWN_RESULT',
					"$journal_file:$line_number",
					'',
					"Result code: $result_code.\nMessages from the test:\n$test_messages"
				);
			}

			# Clear the test point number.
			$test_point = 0;
			$journal_empty = 0;
		}
		elsif ($line_code == 310) {
			# 310 -- Clean Start.
			# Not interesting to us.
		}
		elsif ($line_code == 320) {
			# 320 -- Clean End.
			# Not interesting to us.
		}
		elsif ($line_code == 400) {
			# 400 -- Invocable Component Start.
			# Not interesting to us.
		}
		elsif ($line_code == 410) {
			# 410 -- Invocable Component End.
			# Not interesting to us.
		}
		elsif ($line_code == 510) {
			# 510 -- Error reported by the underlying test suite.
			add_journal_entry(
				$journal_title_data,
				'Test Suite Error',
				"$test_name:$test_point",
				'TS_ERROR',
				'',
				'',
				$line_text
			);
		}
		elsif ($line_code == 520) {
			# 520 -- Message from the underlying test suite.

			# Coalesce consequent rows of spaces in the message
			# to a single space. This improves matching with markers.
			my $message = $line_text;
			$message =~ s/\s+/ /g;

			# Add the message to the accumulated messages.
			$test_messages .= "$message\n";
		}
		elsif ($line_code == 600) {
			# 600 -- Parallel Start.
			# Not interesting to us.
		}
		elsif ($line_code == 620) {
			# 620 -- Parallel End.
			# Not interesting to us.
		}
		elsif ($line_code == 630) {

			# 630 -- Implied Sequential Start.
			# Not interesting to us.
		}
		elsif ($line_code == 640) {
			# 640 -- Implied Sequential End.
			# Not interesting to us.
		}
		elsif ($line_code == 700) {
			# 700 -- Repeat Start.
			# Not interesting to us.
		}
		elsif ($line_code == 720) {
			# 720 -- Repeat End.
			# Not interesting to us.
		}
		elsif ($line_code == 730) {
			# 730 -- Timed Loop Start.
			# Not interesting to us.
		}
		elsif ($line_code == 740) {
			# 740 -- Timed Loop End.
			# Not interesting to us.
		}
		elsif ($line_code == 750) {
			# 750 -- Random Start.
			# Not interesting to us.
		}
		elsif ($line_code == 760) {
			# 760 -- Random End.
			# Not interesting to us.
		}
		elsif ($line_code == 800) {
			# 800 -- Remote Start.
			# Not interesting to us.
		}
		elsif ($line_code == 820) {
			# 820 -- Remote End.
			# Not interesting to us.
		}
		elsif ($line_code == 830) {
			# 830 -- Distributed Start.
			# Not interesting to us.
		}
		elsif ($line_code == 840) {
			# 840 -- Distributed End.
			# Not interesting to us.
		}
		elsif ($line_code == 900) {
			# 900 -- Test Case Controller End.
			# Not interesting to us.
		}
		else {
			# Other lines are notified to the user and otherwise ignored.
			add_journal_entry(
				$journal_title_data,
				'Notice',
				"$test_name:$test_point",
				'J_FORMAT_ERROR',
				"$journal_file:$line_number",
				'',
				"Unknown code number $line_code. Journal line:\n$line"
			);
		}
	}

	if ($journal_empty) {
		$journal_title_data->{'IS_RESULT'} = 0;
	}
	close(JOURNAL);
}

sub sum_stats() {
	my %stats = (
		'PASSES' => 0,
		'FAILS' => 0,
		'WARNS' => 0,
		'FIPS' => 0,
		'NOTICES' => 0,
		'QUELLED' => 0,
		'SKIPPED' => 0
	);
	my $total = 0;
	foreach my $journal (keys %journal_data) {
		if ($journal_data{$journal}->{'IS_RESULT'}) {
			foreach my $st (keys %stats) {
				$stats{$st} += $journal_data{$journal}->{'STATISTICS'}{$st};
				$total += $stats{$st};
			}
		}
	}
	$stats{'TOTAL'} = $total;
	return \%stats;
}

# Returns maximum value from the array.
sub max(@) {
	if (scalar(@_) == 0) {
		return undef;
	}
	my $res = $_[0];
	foreach (@_) {
		$res = $_ if ($res < $_);
	}
	return $res;
}

# Returns array with number of passes, failes and warnings in the test.
# The test is specified via unique HTML anchor.
sub report_testcases_info($) {
	my $journal_anchor = $_[0];
	if (!$journal_anchor) {
		return (-1, -1, -1);
	}
	if (!$journal_title_from_anchor{$journal_anchor}) {
		return (-1, -1, -1);
	}
	my $journal_title = $journal_title_from_anchor{$journal_anchor};
	return ($journal_data{$journal_title}->{'STATISTICS'}->{'PASSES'}, $journal_data{$journal_title}->{'STATISTICS'}->{'FAILS'}, $journal_data{$journal_title}->{'STATISTICS'}->{'WARNS'});
}

my $separator_line = '<img src="images/environment-spacer.gif" width="100%" height="1" style="background-color: white;" alt="" />';
my @verdicts       = ('&nbsp;',  '&nbsp;',  'Success', 'Warnings', 'Failed');
my @verdict_colors = ('#e7e7e7', '#d0d0ff', '#40ff40', '#ffff40',  '#ff6060');

sub report_failures_verdict($$) {
	my ($journal_anchor, $is_root) = @_;

	if (!$journal_anchor) {
		return ($is_root ? 1 : 0); # Nothing
	}

	if (!$journal_title_from_anchor{$journal_anchor}) {
		return ($is_root ? 1 : 0); # Nothing
	}

	my $journal_title = $journal_title_from_anchor{$journal_anchor};

	if (!$journal_data{$journal_title}->{'IS_RESULT'}) {
		return 4; # Failed (No result)
	}

	my $total_stats = $journal_data{$journal_title}->{'STATISTICS'};
	if ($total_stats->{'FAILS'} > 0) {
		return 4; # Failed
	}
	elsif (($total_stats->{'WARNS'} > 0) or ($total_stats->{'FIPS'} > 0)) {
		return 3; # Warning
	}
	else {
		return 2; # Success
	}
}

sub report_failures_cell($$) {
	my ($journal_anchor, $is_root) = @_;
	my $bg_color = $verdict_colors[$is_root ? 1 : 0];

	if (!$journal_anchor) {
		return "<div style=\"background-color: $bg_color;\"><img src=\"images/environment-spacer.gif\" alt=\"\" width=\"5\" height=\"21\" />&nbsp;<br /></div>$separator_line";
	}

	if (!$journal_title_from_anchor{$journal_anchor}) {
		return "<div style=\"background-color: $bg_color;\"><img src=\"images/environment-spacer.gif\" alt=\"\" width=\"5\" height=\"21\" />&nbsp;<br /></div>$separator_line";
	}

	my $journal_title = $journal_title_from_anchor{$journal_anchor};

	my $status = '';
	my $total_stats = $journal_data{$journal_title}->{'STATISTICS'};
	if (!$journal_data{$journal_title}->{'IS_RESULT'}) {
		$bg_color = $verdict_colors[4];
		$status = '<span title="\'No result\' means the tests failed to run, interrupted or crashed during the execution.">No Result<span class="view_srv">&nbsp; <a href="tests_help.pl#No_result" title="Help" target="_blank"><img src="images/question.png" alt="[?]" width="16" height="16" /></a></span></span>';
	}
	elsif ($total_stats->{'FAILS'} > 0) {
		$bg_color = $verdict_colors[4];
		$status = 'Failures: '.$total_stats->{'FAILS'};
	}
	elsif ($total_stats->{'WARNS'} > 0) {
		$bg_color = $verdict_colors[3];
		$status = 'Warnings: '.$total_stats->{'WARNS'};
	}
	else {
		$bg_color = $verdict_colors[2];
		$status = 'Success';
	}

	return "<div style=\"background-color: $bg_color;\"><img src=\"images/environment-spacer.gif\" alt=\"\" width=\"5\" height=\"21\" />$status<img src=\"images/environment-spacer.gif\" alt=\"\" width=\"5\" height=\"21\" /><br /></div>$separator_line";
}

sub report_passes_cell($$) {
	my ($journal_anchor, $is_root) = @_;
	my $bg_color = $verdict_colors[$is_root ? 1 : 0];
	my $passes = (report_testcases_info($journal_anchor))[0];
	if ($passes == -1) {
		return "<div style=\"background-color: $bg_color;\"><img src=\"images/environment-spacer.gif\" alt=\"\" width=\"5\" height=\"21\" />&nbsp;<br /></div>$separator_line";
	}
	else {
		return "<div style=\"background-color: $bg_color;\"><img src=\"images/environment-spacer.gif\" alt=\"\" width=\"5\" height=\"21\" />Passed: $passes<img src=\"images/environment-spacer.gif\" alt=\"\" width=\"5\" height=\"21\" /><br /></div>$separator_line";
	}
}

my $failures_cell = "";
my $passes_cell = "";

sub print_tree_tab($$$$);       # Prototype for recursive function
sub print_tree_tab($$$$) {
	my ($REPORT, $j_tree, $parent_id, $tab) = @_;
	my %sum_my_info = ("passes" => 0, "test_status" => 0, "passes_cell" => "", "failures_cell" => "");
	my @children = sort keys %$j_tree;
	for (my $i = 0; $i <= $#children; $i++) {
		my $elem_name = $children[$i];
		my $elem_id = $elem_name;
		$elem_id =~ s/\s/-/g;
		$elem_id = "$parent_id:$elem_id";
		my $img_f_type;
		my $img_plus_minus;
		my $img_suffix;
		my $tab_new;
		if ($i < $#children) {
			$img_suffix = "d";
			$tab_new = "$tab<img src=\"images/tree_d.png\" alt=\"[ ]\" width=\"21\" height=\"21\" />";
		}
		else {
			$img_suffix = "";
			$tab_new = "$tab<img src=\"images/tree_n.png\" alt=\"[ ]\" width=\"21\" height=\"21\" />";
		}

		if ($j_tree->{$elem_name}{"type"} eq "dir") {
			$img_f_type = "<img src=\"images/tree_dir.png\" alt=\"[D]\" width=\"21\" height=\"21\" />";
			$img_plus_minus = "<img src=\"images/tree_minus_u$img_suffix.png\" alt=\"[-]\" width=\"21\" height=\"21\" id=\"icon$elem_id\" style=\"cursor:pointer;\" onclick=\"javascript:show_hide('$elem_id', '_u$img_suffix');\" />";
		}
		elsif ($j_tree->{$elem_name}{"type"} eq "pkg") {
			$img_f_type = "<img src=\"images/tree_pkg.png\" alt=\"[P]\" width=\"21\" height=\"21\" />";
			if ($j_tree->{$elem_name}{"children"}) {
				$img_plus_minus = "<img src=\"images/tree_minus_u$img_suffix.png\" alt=\"[-]\" width=\"21\" height=\"21\" id=\"icon$elem_id\" style=\"cursor:pointer;\" onclick=\"javascript:show_hide('$elem_id', '_u$img_suffix');\" />";
			}
			else {
				$img_plus_minus = "<img src=\"images/tree_dr$img_suffix.png\" alt=\"[ ]\" width=\"21\" height=\"21\" />";
			}
		}
		else {
			$img_f_type = "<img src=\"images/tree_file.png\" alt=\"[F]\" width=\"21\" height=\"21\" />";
			$img_plus_minus = "<img src=\"images/tree_dr$img_suffix.png\" alt=\"[ ]\" width=\"21\" height=\"21\" />";
		}

		my $elem_link = $elem_name;
		if ($j_tree->{$elem_name}{"journal_anchor"}) {
			$elem_link = "<a href=\"#".$j_tree->{$elem_name}{"journal_anchor"}."\" title=\"Click here to proceed to details\">$elem_link</a>";
			push @journal_anchors_sorted, $j_tree->{$elem_name}{"journal_anchor"};
		}
		print $REPORT "<div style=\"background-color: #e7e7e7; white-space: nowrap;\">$tab$img_plus_minus$img_f_type<b>$elem_link</b><img src=\"images/environment-spacer.gif\" width=\"5\" height=\"21\" alt=\"\" /><br /></div>$separator_line";

		# Calculate passes and failures for the element and its children
		my $local_passes_cell;
		my $local_failures_cell;
		my ($local_test_passes, $local_test_failures, $local_test_warnings) = report_testcases_info($j_tree->{$elem_name}{"journal_anchor"});
		my $local_test_status;
		if ($local_test_passes == -1) {
			$local_passes_cell = "<div style=\"background-color: #e7e7e7;\"><img src=\"images/environment-spacer.gif\" alt=\"\" width=\"5\" height=\"21\" />&nbsp;<br /></div>$separator_line";
			$local_failures_cell = "<div style=\"background-color: #e7e7e7;\"><img src=\"images/environment-spacer.gif\" alt=\"\" width=\"5\" height=\"21\" />&nbsp;<br /></div>$separator_line";
			$local_test_passes = 0;
			$local_test_status = 0;
		}
		else {
			$local_passes_cell = report_passes_cell($j_tree->{$elem_name}{"journal_anchor"}, 0);
			$local_test_status = report_failures_verdict($j_tree->{$elem_name}{"journal_anchor"}, 0);
			$local_failures_cell = report_failures_cell($j_tree->{$elem_name}{"journal_anchor"}, 0);
			$sum_my_info{"passes"} += $local_test_passes;
			$sum_my_info{"test_status"} = max($local_test_status, $sum_my_info{"test_status"});
		}
		if ($j_tree->{$elem_name}{"children"}) {
			print $REPORT "<div id=\"root$elem_id\">\n";
			my %sum_children_info = print_tree_tab($REPORT, $j_tree->{$elem_name}{"children"}, $elem_id, $tab_new);
			print $REPORT "</div>\n";
			$local_test_passes += $sum_children_info{"passes"};
			$sum_my_info{"passes"} += $local_test_passes;
			$local_passes_cell = "<div style=\"display: none;\" id=\"pass-total$elem_id\">\n<div style=\"background-color: #e7e7e7;\"><img src=\"images/environment-spacer.gif\" alt=\"\" width=\"5\" height=\"21\" />Passed: ".$local_test_passes."<img src=\"images/environment-spacer.gif\" alt=\"\" width=\"5\" height=\"21\" /><br /></div>$separator_line</div>\n<div id=\"pass-children$elem_id\">\n".$local_passes_cell.$sum_children_info{"passes_cell"}."</div>\n";
			$sum_my_info{"passes_cell"} .= $local_passes_cell;

			$local_test_status = max($local_test_status, $sum_children_info{"test_status"});
			$sum_my_info{"test_status"} = max($local_test_status, $sum_my_info{"test_status"});
			$local_failures_cell = "<div style=\"display: none;\" id=\"fail-total$elem_id\">\n<div style=\"background-color: ".$verdict_colors[$local_test_status].";\"><img src=\"images/environment-spacer.gif\" alt=\"\" width=\"5\" height=\"21\" />".$verdicts[$local_test_status]."<img src=\"images/environment-spacer.gif\" alt=\"\" width=\"5\" height=\"21\" /><br /></div>$separator_line</div>\n<div id=\"fail-children$elem_id\">\n".$local_failures_cell.$sum_children_info{"failures_cell"}."</div>\n";
			$sum_my_info{"failures_cell"} .= $local_failures_cell;
		}
		else {
			$sum_my_info{"passes_cell"} .= $local_passes_cell;
			$sum_my_info{"failures_cell"} .= $local_failures_cell;
		}
	}
	return %sum_my_info;
}

sub print_tree($$) {
	my ($REPORT, $j_tree) = @_;
	print $REPORT "<table border=\"0\" cellpadding=\"0\" cellspacing=\"0\"><tr><td><img src=\"images/environment-spacer.gif\" width=\"2\" height=\"1\" alt=\"\" /></td><td valign=\"top\">\n";
	foreach my $elem_name (sort keys %$j_tree) {
		my $elem_id = $elem_name;
		$elem_id =~ s/\//:/g;
		$elem_id =~ s/^PKG: /:PKG::/;
		$elem_id =~ s/\s/-/g;
		my $img_f_type;
		my $img_plus_minus;
		if ($j_tree->{$elem_name}{"type"} eq "dir") {
			$img_f_type = "<img src=\"images/tree_dir.png\" alt=\"[D]\" width=\"21\" height=\"21\" />";
			$img_plus_minus = "<img src=\"images/tree_plus.png\" alt=\"[+]\" width=\"21\" height=\"21\" id=\"icon$elem_id\" style=\"cursor:pointer;\" onclick=\"javascript:show_hide('$elem_id', '');\" />";
		}
		elsif ($j_tree->{$elem_name}{"type"} eq "pkg") {
			$img_f_type = "<img src=\"images/tree_pkg.png\" alt=\"[P]\" width=\"21\" height=\"21\" />";
			if ($j_tree->{$elem_name}{"children"}) {
				$img_plus_minus = "<img src=\"images/tree_plus.png\" alt=\"[+]\" width=\"21\" height=\"21\" id=\"icon$elem_id\" style=\"cursor:pointer;\" onclick=\"javascript:show_hide('$elem_id', '');\" />";
			}
			else {
				$img_plus_minus = "<img src=\"images/tree_n.png\" alt=\"[ ]\" width=\"21\" height=\"21\" />";
			}
		}
		else {
			$img_f_type = "<img src=\"images/tree_file.png\" alt=\"[F]\" width=\"21\" height=\"21\" />";
			$img_plus_minus = "<img src=\"images/tree_n.png\" alt=\"[ ]\" width=\"21\" height=\"21\" />";
		}
		my $elem_link = $elem_name;
		if ($j_tree->{$elem_name}{"journal_anchor"}) {
			$elem_link = "<a href=\"#".$j_tree->{$elem_name}{"journal_anchor"}."\" title=\"Click here to proceed to details\">$elem_link</a>";
			push @journal_anchors_sorted, $j_tree->{$elem_name}{"journal_anchor"};
		}
		print $REPORT "<div style=\"background-color: #d0d0ff; white-space: nowrap;\">\n$img_plus_minus$img_f_type<b>$elem_link</b><img src=\"images/environment-spacer.gif\" width=\"5\" height=\"21\" alt=\"\" /><br />\n</div>$separator_line";

		# Calculate passes and failures for the element and its children
		my $local_passes_cell;
		my $local_failures_cell;
		my ($local_test_passes, $local_test_failures, $local_test_warnings) = report_testcases_info($j_tree->{$elem_name}{"journal_anchor"});
		my $local_test_status;
		if ($local_test_passes == -1) {
			$local_passes_cell = "<div style=\"background-color: #d0d0ff;\"><img src=\"images/environment-spacer.gif\" alt=\"\" width=\"5\" height=\"21\" />&nbsp;<br /></div>$separator_line";
			$local_failures_cell = "<div style=\"background-color: #d0d0ff;\"><img src=\"images/environment-spacer.gif\" alt=\"\" width=\"5\" height=\"21\" />&nbsp;<br /></div>$separator_line";
			$local_test_passes = 0;
			$local_test_status = 1;
		}
		else {
			$local_passes_cell = report_passes_cell($j_tree->{$elem_name}{"journal_anchor"}, 1);
			$local_test_status = report_failures_verdict($j_tree->{$elem_name}{"journal_anchor"}, 1);
			$local_failures_cell = report_failures_cell($j_tree->{$elem_name}{"journal_anchor"}, 1);
		}
		if ($j_tree->{$elem_name}{"children"}) {
			print $REPORT "<div id=\"root$elem_id\" style=\"display: none;\">\n";
			my %sum_children_info = print_tree_tab($REPORT, $j_tree->{$elem_name}{"children"}, $elem_id, "<img src=\"images/tree_n.png\" alt=\"[ ]\" width=\"21\" height=\"21\" />");
			print $REPORT "</div>\n";
			$local_test_passes += $sum_children_info{"passes"};
			$local_passes_cell = "<div id=\"pass-total$elem_id\">\n<div style=\"background-color: #d0d0ff;\"><img src=\"images/environment-spacer.gif\" alt=\"\" width=\"5\" height=\"21\" />Passed: ".$local_test_passes."<img src=\"images/environment-spacer.gif\" alt=\"\" width=\"5\" height=\"21\" /><br /></div>$separator_line</div>\n<div style=\"display: none;\" id=\"pass-children$elem_id\">\n".$local_passes_cell.$sum_children_info{"passes_cell"}."</div>\n";
			$local_test_status = max($local_test_status, $sum_children_info{"test_status"});
			$local_failures_cell = "<div id=\"fail-total$elem_id\">\n<div style=\"background-color: ".$verdict_colors[$local_test_status].";\"><img src=\"images/environment-spacer.gif\" alt=\"\" width=\"5\" height=\"21\" />".$verdicts[$local_test_status]."<img src=\"images/environment-spacer.gif\" alt=\"\" width=\"5\" height=\"21\" /><br /></div>$separator_line</div>\n<div style=\"display: none;\" id=\"fail-children$elem_id\">\n".$local_failures_cell.$sum_children_info{"failures_cell"}."</div>\n";
		}
		$passes_cell .= $local_passes_cell;
		$failures_cell .= $local_failures_cell;
	}
	print $REPORT <<DATA;
</td>
<td><img src="images/environment-spacer.gif" width="1" height="1" alt="" /></td>
<td valign="top" style="white-space: nowrap;">
$failures_cell
</td>
<td><img src="images/environment-spacer.gif" width="1" height="1" alt="" /></td>
<td valign="top" style="white-space: nowrap;">
$passes_cell
</td>
</tr></table>
DATA
}

sub host_info() {
	return <<DATA;
<h3>Host Info</h3>

<table class="wikitable">
  <tr>
    <th style="text-align: left;">Host Name</th>
    <td>$host_name</td>
  </tr>
  <tr>
    <th style="text-align: left;">Host OS</th>
    <td>$host_os</td>
  </tr>
  <tr>
    <th style="text-align: left;">Host Machine</th>
    <td>$host_machine</td>
  </tr>
  <tr>
    <th style="text-align: left;">Test Architecture</th>
    <td>$test_arch</td>
  </tr>
  <tr>
    <th style="text-align: left;">Host Kernel</th>
    <td>$host_kernel</td>
  </tr>
</table>
The complete console output log is available <a href="/get.pl$result_path/$result_dir/test_logs/tests_output" class="view_srv" target="_blank">here</a><a href="test_logs/tests_output" class="view_as_is" target="_blank">here</a>.
<br />
<br />

DATA
}

# Creates the HTML report
sub compile_report() {
	my $REPORT;
	if (!open($REPORT, ">$result_path/$result_dir/test_log.html")) {
		print "Couldn't open file '$result_path/$result_dir/test_log.html' for writing: $!";
		return;
	}

	my $bugreport_email = 'lsb-appcheck-support@linuxfoundation.org';
	
	# Print Title
	print $REPORT <<DATA;
<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">
<html xmlns="http://www.w3.org/1999/xhtml" xml:lang="en" lang="en" dir="ltr">
<head>
  <title>Moblin $moblin_version Test Results for $application_name on $test_arch</title>
  <meta http-equiv="Content-Type" content="text/html; charset=utf-8" />
  <style type="text/css">
    .view_srv {
        display: none;
    }
    table {
      empty-cells: show;
    }
    table.wikitable, table.wikitable-sep {
      margin: 1em 1em 1em 0;
      background: #f9f9f9;
    }
    table.wikitable {
      border: 1px #aaaaaa solid;
      border-collapse: collapse;
    }
    table.wikitable th, table.wikitable td {
      border: 1px #aaaaaa solid;
      padding: 0.2em;
      padding-left: 0.4em;
      padding-right: 0.4em;
    }
    table.wikitable-sep th, table.wikitable-sep td {
      border-bottom: 1px #aaaaaa solid;
      border-right: 1px #aaaaaa solid;
      padding: 0.2em;
      padding-left: 0.4em;
      padding-right: 0.4em;
    }
    table.wikitable-sep th.btop, table.wikitable-sep td.btop {
      border-top: 1px #aaaaaa solid;
    }
    table.wikitable-sep th.bleft, table.wikitable-sep td.bleft {
      border-left: 1px #aaaaaa solid;
    }
    table.wikitable th.left {
      background: #f2f2f2;
      text-align: left;
      padding-left: 0.4em;
      padding-right: 0.4em;
    }
    table.wikitable th, table.wikitable-sep th {
      background: #f2f2f2;
      text-align: center;
      padding-left: 0.4em;
      padding-right: 0.4em;
    }
  </style>
</head>

<body>

<script language="javascript" type="text/javascript">
// <![CDATA[
// Preload some images that are not initially visible
var preload_image_names = new Array('tree_d.png', 'tree_dir.png', 'tree_dr.png', 'tree_drd.png', 'tree_file.png', 'tree_minus.png', 'tree_minus_u.png', 'tree_minus_ud.png', 'tree_n.png', 'tree_pkg.png', 'tree_plus.png', 'tree_plus_u.png', 'tree_plus_ud.png', 'report_plus.png', 'report_minus.png');
var preload_images_arr = new Array();
for (var i=0; i<preload_image_names.length; ++i) {
	preload_images_arr[i] = new Image;
	preload_images_arr[i].src = 'images/' + preload_image_names[i];
}

function show_hide(id, icon_suffix) {
	var elem = document.getElementById('root' + id);
	var icon = document.getElementById('icon' + id);
	var passTotal = document.getElementById('pass-total' + id);
	var passChildren = document.getElementById('pass-children' + id);
	var failTotal = document.getElementById('fail-total' + id);
	var failChildren = document.getElementById('fail-children' + id);
	if (elem && icon && passTotal && passChildren && failTotal && failChildren) {
		var to_show = (elem.style.display == 'none');
		if (to_show) {
			elem.style.display = '';
			passTotal.style.display = 'none';
			passChildren.style.display = '';
			failTotal.style.display = 'none';
			failChildren.style.display = '';
			icon.setAttribute('src', 'images/tree_minus' + icon_suffix + '.png');
			icon.setAttribute('alt', '[-]');
		}
		else {
			elem.style.display = 'none';
			passTotal.style.display = '';
			passChildren.style.display = 'none';
			failTotal.style.display = '';
			failChildren.style.display = 'none';
			icon.setAttribute('src', 'images/tree_plus' + icon_suffix + '.png');
			icon.setAttribute('alt', '[+]');
		}
	}
}

function bug_report(journal_path, test_app, test_case) {
	if (confirm('A new E-mail to $bugreport_email will now be created.\\nPlease specify all the necessary details and attach the journal files:\\n\\n' + journal_path + '\\n\\nContinue?')) {
		document.location = 'mailto:$bugreport_email?Subject=AppChecker:%20Problem%20Report%20(' + test_case + ')&Body=File tested: ' + test_app + '%0A%0A';
	}
}

function submit_problem_info(journal_path, test_app, test_case) {
	if (confirm('A new E-mail to $bugreport_email will now be created.\\nPlease specify all the necessary details and attach the journal files:\\n\\n' + journal_path + '\\n\\nContinue?')) {
		document.location = 'mailto:$bugreport_email?Subject=AppChecker:%20Submit%20a%20Comment%20(' + test_case + ')&Body=File tested: ' + test_app + '%0A%0AComments:%0A';
	}
}

function show_hide_msg(img, id) {
	var elem = document.getElementById(id);
	if (!elem)
		return;
	var collapsed = (elem.style.display == 'none');
	if (collapsed) {
		elem.style.display = '';
		img.setAttribute('src', 'images/report_minus.png');
	}
	else {
		elem.style.display = 'none';
		img.setAttribute('src', 'images/report_plus.png');
	}
}

function switch_group(name) {
	var elem_cat = document.getElementById('by-category');
	var elem_comp = document.getElementById('by-component');
	if (elem_cat && elem_comp) {
		if (name == 'component') {
			elem_cat.style.display = 'none';
			elem_comp.style.display = '';
		}
		else {
			elem_cat.style.display = '';
			elem_comp.style.display = 'none';
		}
	}
}
// ]]>
</script>

<h1 class="view_as_is">Moblin $moblin_version Test Results for <font color="blue">$application_name</font> on <font color="blue">$test_arch</font></h1>

<!-- Test Report Header -->
<br />

<table class="bartable">
  <tr>
    <td align="right"><b>Group Test Results by:</b></td>
    <td align="right">
      <select name="moblin_filter_libs" id="moblin_filter_libs" onchange="javascript:switch_group(this.value);" style="width: 12.5em;">
        <option value="category" selected="selected">Category</option>
        <option value="component">Component</option>
      </select>
    </td>
  </tr>
</table>
<br />

<div id="by-component" style="display: none;">
<h3>Test Summary</h3>
DATA

	# Print the index tree
	print_tree($REPORT, \%journal_tree);
	print $REPORT "<br />\n\n";
	print $REPORT host_info();

	print $REPORT <<DATA;
<h1>Detailed Reports by Components</h1>
DATA

	# Test journals
	my $first_journal = 1;
	foreach my $journal_anchor (@journal_anchors_sorted) {
		my $journal_title;
		if ($journal_title_from_anchor{$journal_anchor}) {
			$journal_title = $journal_title_from_anchor{$journal_anchor};
		}
		else {
			next;
		}

		# User-friendly file path
		my $journal_title_desc = $journal_title;
		$journal_title_desc =~ s/^\|//;
		$journal_title_desc =~ s/\|$//;
		$journal_title_desc =~ s/\|/ =&gt; /g;
		if ($first_journal) {
			$first_journal = 0;
		}
		else {
			print $REPORT "<hr />\n<br />\n";
		}
		print $REPORT "<h2><a name=\"$journal_anchor\"></a><font color=\"#336699\">$journal_title_desc</font></h2>\n";

		print $REPORT generate_journal_report($journal_anchor, $journal_title, $journal_title_desc);
	}
	print $REPORT <<DATA;
</div>

<div id="by-category">
<h3>Test Summary</h3>
DATA
	# Print the category list
	my @filtered_list = ();
	print $REPORT <<DATA;
<table cellpadding="1" cellspacing="1" border="0">
DATA
	foreach my $cat (@categories_sort) {
		next if (!defined($cat_problems{$cat}) or (scalar(@{$cat_problems{$cat}->{'PROBLEMS'}}) == 0));
		push @filtered_list, $cat;
		my $lc_cat = lc($cat);
		$lc_cat =~ s/[^a-z0-9\-_:.]/-/g;
		my $num_fails = $cat_problems{$cat}->{'FAILS'};
		my $color = $verdict_colors[($num_fails ? 4 : 0)];
		print $REPORT <<DATA;
  <tr>
    <td style="background-color: $verdict_colors[1];"><img src="images/tree_file.png" alt="" width="21" height="21" /><b><a href="#$lc_cat" title="Click here to proceed to details">$cat</a></b></td>
    <td style="background-color: $color;">Failures: $num_fails</td>
  </tr>
DATA
	}
	print $REPORT <<DATA;
</table>
<br />
DATA
	print $REPORT host_info();

	print $REPORT <<DATA;
<h1>Detailed Reports by Category</h1>
DATA

	# Print the problems
	foreach my $cat (@filtered_list) {
		my $lc_cat = lc($cat);
		$lc_cat =~ s/[^a-z0-9\-_:.]/-/g;
		my $elem_col_name = $categories{$cat}->{'ELEM_COL_NAME'};
		print $REPORT <<DATA;
<h2><a name="$lc_cat"></a><font color="#336699">$cat</font></h2>
$categories{$cat}->{'DESCRIPTION'}
<table class="wikitable-sep" cellpadding="0" cellspacing="0">
  <tr>
    <th class="btop bleft" style="width: 17px;"></th>
    <th class="btop" style="width: 25em;">Component</th>
    <th class="btop" style="width: 5em;">Type</th>
    <th class="btop" style="width: 10em;">$elem_col_name</th>
    <th class="btop" style="width: 25em;">Comments</th>
  </tr>
DATA
		# Sort problems list by element
		my @problems_list_sorted = sort {
			$journal_data{$a->{'JOURNAL_TITLE'}}->{'PROBLEMS_LIST'}->[$a->{'PROBLEM_IDX'}]->{'ELEMENT'}
			cmp
			$journal_data{$b->{'JOURNAL_TITLE'}}->{'PROBLEMS_LIST'}->[$b->{'PROBLEM_IDX'}]->{'ELEMENT'}
		} @{$cat_problems{$cat}->{'PROBLEMS'}};
		foreach (my $i = 0; $i < scalar(@problems_list_sorted); ++$i) {
			my $journal_title = $problems_list_sorted[$i]->{'JOURNAL_TITLE'};
			my $journal_title_desc = $journal_title;
			$journal_title_desc =~ s/\'/\\\'/g;
			my $id = "$lc_cat-$i";
			my $idx = $problems_list_sorted[$i]->{'PROBLEM_IDX'};
			my $problem = $journal_data{$journal_title}->{'PROBLEMS_LIST'}->[$idx];
			my $severity = $problem->{'SEVERITY'};
			my $severity_color = '';
			my $subcategory = $subcategories{$problem->{'SUBCATEGORY'}};
			if (($severity =~ m/Error/i) or
				($severity =~ m/Failed/i) or
				($severity =~ m/Unresolved/i) or
				($severity =~ m/Uninitiated/i) or
				($severity =~ m/Unreported/i) or
				($severity =~ m/Time Expired/i)) {
				$severity_color = " background-color: $verdict_colors[4];";
			}
			elsif ($severity =~ m/Warning/i) {
				$severity_color = " background-color: $verdict_colors[3];";
			}
			elsif ($severity =~ m/FIP/i) {
				$severity_color = ' background-color: #a0a0ff;';
				$severity = "<span title=\"'FIP' stands for 'Further Information Provided'.\">$severity<span class=\"view_srv\">&nbsp; <a href=\"tests_help.pl#FIP\" title=\"Help\" target=\"_blank\"><img src=\"images/question.png\" alt=\"[?]\" width=\"16\" height=\"16\" /></a></span></span>";
			}
			my $element = $problem->{'ELEMENT'};
			if ($subcategory->{'ELEM_LINK'} and ($element ne '') and ($element ne '...')) {
				my $link_element = $element;
				$link_element =~ s/\+/%2B/g;
				$element = "<a href=\"$subcategory->{'ELEM_LINK'}$link_element\" target=\"_blank\">$element</a>";
			}
			my $comment = ($problem->{'COMMENT'} ? '<td>'.$problem->{'COMMENT'}.'</td>' : '<td style="color: #a0a0a0;">No information</td>');
			my $test_msg = $problem->{'DETAILS'};
			$test_msg =~ s/&/&amp;/g;
			$test_msg =~ s/</&lt;/g;
			$test_msg =~ s/>/&gt;/g;
			$test_msg =~ s/\n/<br \/>/g;
			my $component = $journal_title;
			$component =~ s/^.*\.rpm\|([^|]+)/RPM:$1/;
			$component =~ s/^.*\.deb\|([^|]+)/DEB:$1/;
			$component =~ s/^.*\.(tar)\|([^|]+)/TAR:$2/;
			$component =~ s/^.*\.(tar\.gz|tgz)\|([^|]+)/TGZ:$2/;
			$component =~ s/^.*\.(tar\.bz2|tbz)\|([^|]+)/TBZ:$2/;
			$component =~ s/^PKG:.*\|([^|]+)/PKG:$1/;
			$component =~ s/\|//g;
			$journal_title =~ s/\|$//;
			$journal_title =~ s/\|/ =&gt; /g;
			my $test_case = $problem->{'TEST_CASE'};
			$test_case =~ s/\'/\\\'/g;
			print $REPORT <<DATA;
  <tr>
    <td class="bleft" rowspan="2" style="text-align: center; vertical-align: top;"><img src="images/report_plus.png" alt="[+]" width="17" height="17" style="cursor: pointer;" onclick="javascript:show_hide_msg(this, '$id');" /></td>
    <td style="white-space: nowrap;" title="$journal_title">$component</td>
    <td style="text-align: center;$severity_color"><b>$severity</b></td>
    <td style="white-space: nowrap;">$element</td>
    $comment
  </tr>
  <tr>
    <td colspan="4" id="$id" style="display: none;">
      $subcategory->{'MESSAGE'}<br/><div style='font-family:monospace; padding-left: 3em;'>$test_msg</div><div align="right"><input type="button" value="Submit Your Comment" style="width: 12em;" onclick="javascript:submit_problem_info('$result_path/$result_dir/$result_archive', '$journal_title_desc', '$test_case');" title="Press this button to submit additional information about this problem." />&nbsp;<input type="button" value="Report Bug in the Test" style="width: 12em;" onclick="javascript:bug_report('$result_path/$result_dir/$result_archive', '$journal_title_desc', '$test_case');" title="Press this button if you beleive the problem is in the test, not in your system." /></div>
    </td>
  </tr>
DATA
		}
		print $REPORT "</table>\n";
	}

	print $REPORT <<DATA;
</div>
</body>
</html>
DATA
	close($REPORT);
}

sub generate_journal_report($$$) {
	my ($journal_anchor, $journal_title, $journal_title_desc) = @_;
	my $journal_title_data = $journal_data{$journal_title};
	
	if (!defined($journal_title_data->{'IS_RESULT'})) {
		return <<DATA;
<h3 style="color: #a00000;">No result.</h3>
Journal file for <b>$journal_title_desc</b> could not be opened. This usually means that the corresponding test failed to run, interrupted or crashed during the execution.
DATA
	}

	my $stat_data = $journal_title_data->{'STATISTICS'};
	my $problems_list = $journal_title_data->{'PROBLEMS_LIST'};
	my $journal_file = $journal_title_data->{'JOURNAL_FILE'};
	my $verdict_idx = report_failures_verdict($journal_anchor, 0);
	my $fail_color = ($stat_data->{'FAILS'} ? ' background-color: '.$verdict_colors[4] : '');
	my $warn_color = ($stat_data->{'WARNS'} ? ' background-color: '.$verdict_colors[3] : '');
	my $fip_color = ($stat_data->{'FIPS'} ? ' background-color: '.$verdict_colors[3] : '');
	my $res = '';
	$res = <<DATA;
<h3>Test Statictics</h3>

<table cellpadding="0" cellspacing="0" border="0">
  <tr>
    <td style="vertical-align: top;">
      <table class="wikitable">
        <tr>
          <th style="text-align: left; width: 7em;">Passed tests</th>
          <td style="text-align: right; width: 4em;">$stat_data->{'PASSES'}</td>
        </tr>
        <tr>
          <th style="text-align: left;">Failed tests</th>
          <td style="text-align: right;$fail_color">$stat_data->{'FAILS'}</td>
        </tr>
DATA
	if ($stat_data->{'WARNS'}) {
		$res .= <<DATA;
        <tr>
          <th style="text-align: left;">Warnings</th>
          <td style="text-align: right;$warn_color">$stat_data->{'WARNS'}</td>
        </tr>
DATA
	}
	if ($stat_data->{'FIPS'}) {
		$res .= <<DATA;
        <tr>
          <th style="text-align: left;">FIPs</th>
          <td style="text-align: right;$fip_color">$stat_data->{'FIPS'}</td>
        </tr>
DATA
	}
	$res .= <<DATA;
        <tr>
          <th style="text-align: left;">Verdict</th>
          <td style="text-align: right; background-color: $verdict_colors[$verdict_idx];">$verdicts[$verdict_idx]</td>
        </tr>
      </table>
    </td>
    <td width="20"></td>
    <td style="vertical-align: top;">
      <table class="wikitable">
        <tr>
          <th style="text-align: left;">Test Tool</th>
          <td style="text-align: right;">$journal_title_data->{'TEST_TOOL'} (<a href="/get.pl$result_path/$result_dir/$journal_title_data->{'JOURNAL_FILE'}" class="view_srv" target="_blank">open journal file</a><a href="$journal_title_data->{'JOURNAL_FILE'}" class="view_as_is" target="_blank">open journal file</a>)</td>
        </tr>
        <tr>
          <th style="text-align: left;">Test Start Time</th>
          <td style="text-align: right;">$journal_title_data->{'START_TIME'}</td>
        </tr>
      </table>
    </td>
  </tr>
</table>

DATA
	if (!$journal_title_data->{'IS_RESULT'}) {
		$res .= <<DATA;
<h3 style="color: #a00000;">No result.</h3>
Journal file for <b>$journal_title_desc</b> is empty or contains no test cases. This usually means that the corresponding test failed to run, interrupted or crashed during the execution.
DATA
	}
	elsif (scalar(@$problems_list) == 0) {
		$res .= "<h3>No problems detected</h3>\n";
	}
	else {
		$res .= <<DATA;
<h3>List of Problems</h3>

<table class="wikitable-sep" cellpadding="0" cellspacing="0">
  <tr>
    <th class="btop bleft" style="width: 17px;"></th>
    <th class="btop" style="width: 16em;">Category</th>
    <th class="btop" style="width: 5em;">Type</th>
    <th class="btop" style="width: 10em;">Element</th>
    <th class="btop" style="width: 25em;">Comments</th>
  </tr>
DATA
		$journal_title_desc =~ s/\'/\\\'/g;
		foreach (my $i = 0; $i < scalar(@$problems_list); ++$i) {
			my $id = "$journal_anchor-$i";
			my $severity = $problems_list->[$i]->{'SEVERITY'};
			my $severity_color = '';
			my $subcategory = $subcategories{$problems_list->[$i]->{'SUBCATEGORY'}};
			my $is_fail = 0;
			if (($severity =~ m/Error/i) or
				($severity =~ m/Failed/i) or
				($severity =~ m/Unresolved/i) or
				($severity =~ m/Uninitiated/i) or
				($severity =~ m/Unreported/i) or
				($severity =~ m/Time Expired/i)) {
				$severity_color = " background-color: $verdict_colors[4];";
				$is_fail = 1;
			}
			elsif ($severity =~ m/Warning/i) {
				$severity_color = " background-color: $verdict_colors[3];";
			}
			elsif ($severity =~ m/FIP/i) {
				$severity_color = ' background-color: #a0a0ff;';
				$severity = "<span title=\"'FIP' stands for 'Further Information Provided'.\">$severity<span class=\"view_srv\">&nbsp; <a href=\"tests_help.pl#FIP\" title=\"Help\" target=\"_blank\"><img src=\"images/question.png\" alt=\"[?]\" width=\"16\" height=\"16\" /></a></span></span>";
			}
			my $element = $problems_list->[$i]->{'ELEMENT'};
			if ($subcategory->{'ELEM_LINK'} and ($element ne '') and ($element ne '...')) {
				$element = "<a href=\"$subcategory->{'ELEM_LINK'}$element\" target=\"_blank\">$element</a>";
			}
			my $comment = ($problems_list->[$i]->{'COMMENT'} ? '<td>'.$problems_list->[$i]->{'COMMENT'}.'</td>' : '<td style="color: #a0a0a0;">No information</td>');
			my $category = $subcategory->{'CATEGORY'};
			my $test_msg = $problems_list->[$i]->{'DETAILS'};
			$test_msg =~ s/&/&amp;/g;
			$test_msg =~ s/</&lt;/g;
			$test_msg =~ s/>/&gt;/g;
			$test_msg =~ s/\n/<br \/>/g;
			my $test_case = $problems_list->[$i]->{'TEST_CASE'};
			$test_case =~ s/\'/\\\'/g;
			$res .= <<DATA;
  <tr>
    <td class="bleft" rowspan="2" style="text-align: center; vertical-align: top;"><img src="images/report_plus.png" alt="[+]" width="17" height="17" style="cursor: pointer;" onclick="javascript:show_hide_msg(this, '$id');" /></td>
    <td style="white-space: nowrap;">$category</td>
    <td style="text-align: center;$severity_color"><b>$severity</b></td>
    <td style="white-space: nowrap;">$element</td>
    $comment
  </tr>
  <tr>
    <td colspan="4" id="$id" style="display: none;">
      $subcategory->{'MESSAGE'}<br/><div style='font-family:monospace; padding-left: 3em;'>$test_msg</div><div align="right"><input type="button" value="Submit Your Comment" style="width: 12em;" onclick="javascript:submit_problem_info('$result_path/$result_dir/$result_archive', '$journal_title_desc', '$test_case');" title="Press this button to submit additional information about this problem." />&nbsp;<input type="button" value="Report Bug in the Test" style="width: 12em;" onclick="javascript:bug_report('$result_path/$result_dir/$result_archive', '$journal_title_desc', '$test_case');" title="Press this button if you beleive the problem is in the test, not in your system." /></div>
    </td>
  </tr>
DATA
			# Add problem into problems-in-category list
			if (!defined($cat_problems{$category})) {
				$cat_problems{$category}->{'FAILS'} = 0;
				$cat_problems{$category}->{'PROBLEMS'} = [];
			}
			push @{$cat_problems{$category}->{'PROBLEMS'}}, {
				'JOURNAL_TITLE' => $journal_title,
				'PROBLEM_IDX' => $i
			};
			$cat_problems{$category}->{'FAILS'} += $is_fail;
		}
		$res .= "</table>\n";
	}
	return $res;
}

################################################################################
# Start execution

parse_cmd_line();
read_host_info();
generate_report();
